import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string


class HumanoidBodyVariant(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        self.model_part = Ref(stream, self.version)
        self.ability_pose_deformer = Ref(stream, self.version)
        unk_refs_1_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_1: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_1_count)]
        unk_refs_2_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_2: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_2_count)]
        unk_refs_3_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_3: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_3_count)]
        self.unk_ref_4 = Ref(stream, self.version)
        self.unk_int_5 = struct.unpack('<I', stream.read(4))[0]
        self.unk_ref_6 = Ref(stream, self.version)
        self.unk_int_7 = struct.unpack('<I', stream.read(4))[0]
        self.unk_float_8 = struct.unpack('<f', stream.read(4))[0]
        unk_refs_9_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_9: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_9_count)]
        unk_strings_10_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_strings_10: List[str] = [parse_hashed_string(stream) for _ in range(unk_strings_10_count)]

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



import struct
from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion

from pydecima.resources.Resource import Resource
from pydecima.enums.ETextLanguages import ETextLanguages


class LocalizedTextResource(Resource):
    @staticmethod
    def read_fixed_string(stream: BinaryIO):
        size = struct.unpack('<H', stream.read(2))[0]
        return stream.read(size).decode('UTF8')

    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.language = [LocalizedTextResource.read_fixed_string(stream) for _ in ETextLanguages]

    def __str__(self):
        return self.language[ETextLanguages.English].strip()

    def __repr__(self):
        return self.language[ETextLanguages.English].__repr__()



import struct
from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref


class LocalizedAnimationResource(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.unk_1 = struct.unpack('<I', stream.read(4))[0]
        self.unk_2 = struct.unpack('<I', stream.read(4))[0]
        self.unk_3 = Ref(stream, self.version)



import struct
from codecs import iterdecode
from functools import partial
from itertools import islice
from typing import BinaryIO

from pydecima.resources.structs.HashedString import HashedString


def parse_hashed_string(stream: BinaryIO, include_hash=False):
    size = struct.unpack('<I', stream.read(4))[0]
    if size > 0:
        string_hash = stream.read(4)
    else:
        string_hash = bytes()
    if include_hash:
        return HashedString(string_hash, stream.read(size).decode('ASCII'))
    else:
        return stream.read(size).decode('ASCII')


def read_utf16_chars(stream: BinaryIO, length):
    # Read one byte at a time
    binary_chunks = iter(partial(stream.read, 1), "")
    # Convert bytes into unicode code points
    decoder = iterdecode(binary_chunks, 'utf_16_le')
    return str().join(islice(decoder, length))



from .pc_type_map import pc_type_map
from .ps4_type_map import ps4_type_map
from .dspc_type_map import dspc_type_map
from pydecima.enums.DecimaVersion import DecimaVersion


def get_type_map(version: DecimaVersion):
    if version == DecimaVersion.HZDPC:
        return pc_type_map
    elif version == DecimaVersion.HZDPS4:
        return ps4_type_map
    elif version == DecimaVersion.DSPC:
        return dspc_type_map
    else:
        raise Exception(f"No matching type map found for {version}")



import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.LootSlot import LootSlot
from pydecima.resources.InventoryItemComponentResource import InventoryItemComponentResource
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string


class InventoryLootPackageComponentResource(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        self.unk_ref = Ref(stream, self.version)
        assert (self.unk_ref.type == 0), "unk_ref populated"
        ref_count = struct.unpack('<I', stream.read(4))[0]
        self.loot_slot: List[Ref[LootSlot]] = [Ref(stream, self.version) for _ in range(ref_count)]
        self.inventory_item_component: Ref[InventoryItemComponentResource] = Ref(stream, self.version)

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
import struct
from pydecima.resources.Resource import Resource

# TODO: Unfinished


class CollisionTrigger(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        start_pos = stream.tell()
        Resource.__init__(self, stream, version)
        stream.seek(start_pos + 12)
        self.unk = struct.unpack('<I', stream.read(4))[0]
        self.uuid = stream.read(16)
        stream.seek(start_pos)
        self.data = stream.read(self.size + 12)



from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
import struct
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string, read_utf16_chars


class CreditsColumn(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        name_len = struct.unpack('<I', stream.read(4))[0]
        self.credits_name = read_utf16_chars(stream, name_len)
        self.style = Ref(stream, self.version)
        self.style_2 = Ref(stream, self.version)
        self.unk = Ref(stream, self.version)
        assert self.unk.type == 0

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.ImageStruct import ImageStruct
from pydecima._utils import parse_hashed_string


class Texture(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        start_pos = stream.tell()
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        if self.size - (stream.tell() - start_pos) > 0:
            self.image_data = ImageStruct(stream, version)

    def __str__(self):
        streamed = hasattr(self.image_data, "size_of_stream") and self.image_data.size_of_stream > 0
        return f'{self.type}: {self.name}, {self.image_data.width}x{self.image_data.height}, ' + \
               f'{self.image_data.image_format.name}, {"streamed " if streamed else ""}' + \
               f'@{hex(self.image_data.stream_start) if streamed else "internal"}'



import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref


class ObjectCollection(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        objects_count = struct.unpack('<I', stream.read(4))[0]
        self.objects: List[Ref] = [Ref(stream, self.version) for _ in range(objects_count)]



import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.HumanoidBodyVariant import HumanoidBodyVariant
from pydecima.resources.HumanoidBodyVariantGroup import HumanoidBodyVariantGroup
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string


class SpawnSetup(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        self.graph_condition = Ref(stream, self.version)
        self.impostor = Ref(stream, self.version)
        self.humanoid = Ref(stream, self.version)
        self.graph_program = Ref(stream, self.version)
        self.faction = Ref(stream, self.version)
        unk_refs_1_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_1: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_1_count)]
        unk_refs_2_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_2: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_2_count)]
        self.unk_ints = struct.unpack('<iii', stream.read(12))
        self.unk_byte = struct.unpack('<b', stream.read(1))
        unk_refs_3_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_refs_3: List[Ref] = [Ref(stream, self.version) for _ in range(unk_refs_3_count)]
        self.inventory_collection = Ref(stream, self.version)
        self.combat_behavior = Ref(stream, self.version)
        self.humanoid_body_variant: [Ref[HumanoidBodyVariant], Ref[HumanoidBodyVariantGroup]] =\
            Ref(stream, self.version)
        self.combat_properties = Ref(stream, self.version)
        self.combat_properties_facts = Ref(stream, self.version)

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



import struct
from pydecima.enums.EWaveDataEncoding import EWaveDataEncoding
from pydecima.enums.EWaveDataEncodingQuality import EWaveDataEncodingQuality
from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima._utils import parse_hashed_string


class WaveResource(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        start_pos = stream.tell()
        Resource.__init__(self, stream, version)
        stream.seek(start_pos + 12)
        self.encoding_quality = EWaveDataEncodingQuality(struct.unpack('<I', stream.read(4))[0])
        self.unk_bytes1 = struct.unpack('<bb', stream.read(2))
        self.uuid = stream.read(16)
        self.name = parse_hashed_string(stream)
        self.size_without_stream = struct.unpack('<I', stream.read(4))[0]
        if self.size_without_stream > 0:
            self.sound = stream.read(self.size_without_stream)
        self.size_with_stream = struct.unpack('<I', stream.read(4))[0]
        self.sample_rate = struct.unpack('<I', stream.read(4))[0]
        self.channels = struct.unpack('<b', stream.read(1))[0]
        self.encoding = EWaveDataEncoding(struct.unpack('<I', stream.read(4))[0])
        self.unk_short4 = struct.unpack('<H', stream.read(2))[0]
        self.bit_rate = struct.unpack('<I', stream.read(4))[0]
        self.unk_short5 = struct.unpack('<H', stream.read(2))[0]
        self.unk_short6 = struct.unpack('<H', stream.read(2))[0]
        self.unk_bytes7 = struct.unpack('<bbbbbb', stream.read(6))
        if self.size_with_stream != self.size_without_stream:
            cache_len = struct.unpack('<I', stream.read(4))[0]
            self.cache_string = stream.read(cache_len).decode('ASCII')
            self.unk_ints8 = struct.unpack('<IIII', stream.read(16))

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



import struct
from typing import BinaryIO
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.FocusScannedInfo import FocusScannedInfo
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string


class FocusTargetComponentResource(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        self.boolean_fact = Ref(stream, self.version)
        self.unk_short_1 = struct.unpack('<h', stream.read(2))[0]
        self.unk_short_2 = struct.unpack('<h', stream.read(2))[0]
        self.unk_int_1 = struct.unpack('<I', stream.read(4))[0]
        assert self.unk_int_1 == 0
        self.unk_float = struct.unpack('<f', stream.read(4))[0]
        self.unk_byte_1 = struct.unpack('<b', stream.read(1))[0]
        assert self.unk_byte_1 == 0
        self.focus_scanned_info: Ref[FocusScannedInfo] = Ref(stream, self.version)
        unk_ref_count = struct.unpack('<I', stream.read(4))[0]
        self.unk_ref_list = [Ref(stream, self.version) for _ in range(unk_ref_count)]
        self.unk_byte_2 = struct.unpack('<b', stream.read(1))[0]
        self.focus_tracking_path_entity = Ref(stream, self.version)

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima.resources.structs.HashedString import HashedString
from pydecima._utils import parse_hashed_string


class PrefetchList(Resource):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        paths_count = struct.unpack('<I', stream.read(4))[0]
        self.paths: List[HashedString] = [parse_hashed_string(stream, True) for _ in range(paths_count)]
        sizes_count = struct.unpack('<I', stream.read(4))[0]
        assert (paths_count == sizes_count), "differing number of paths and sizes"
        self.sizes = [struct.unpack('<I', stream.read(4))[0] for _ in range(sizes_count)]
        indices_count = struct.unpack('<I', stream.read(4))[0]
        self.indices = [struct.unpack('<I', stream.read(4))[0] for _ in range(indices_count)]



import struct
from typing import BinaryIO, List
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.enums.ESentenceGroupType import ESentenceGroupType
from pydecima.resources.Resource import Resource
from pydecima.resources.SentenceResource import SentenceResource
from pydecima.resources.structs.Ref import Ref
from pydecima._utils import parse_hashed_string


class SentenceGroupResource(Resource):

    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        Resource.__init__(self, stream, version)
        self.name = parse_hashed_string(stream)
        sentence_type = struct.unpack('<I', stream.read(4))[0]
        self.sentence_type = ESentenceGroupType(sentence_type)
        sentences_count = struct.unpack('<I', stream.read(4))[0]
        self.sentences: List[Ref[SentenceResource]] = [Ref(stream, self.version) for _ in range(sentences_count)]

    def __str__(self):
        return '{}: {}'.format(self.type, self.name)



import binascii
import os
import struct
from typing import TypeVar, Generic, BinaryIO, Dict

import pydecima
from pydecima.enums.DecimaVersion import DecimaVersion
from pydecima.resources.Resource import Resource
from pydecima._utils import parse_hashed_string

T = TypeVar('T')


class Ref(Generic[T]):
    def __init__(self, stream: BinaryIO, version: DecimaVersion):
        self.version = version
        self.type: int = struct.unpack('<B', stream.read(1))[0]
        if self.type > 0:
            self.hash: bytes = stream.read(16)
        if self.type in [2, 3]:
            self.path: str = parse_hashed_string(stream)

    def follow(self, resource_dict: Dict[bytes, Resource]) -> T:
        if self.type == 0:
            return None
        if self.hash in resource_dict:
            return resource_dict[self.hash]
        elif hasattr(self, 'path'):
            full_path = os.path.join(pydecima.reader.game_root, self.path) + '.core'
            pydecima.reader.read_objects(full_path, resource_dict)
            if self.hash in resource_dict:
                return resource_dict[self.hash]
        raise Exception('Resource not in list: {}'.format(self.__str__()))

    def __str__(self):
        ret = 'Type {} Ref: '.format(self.type)
        if hasattr(self, 'hash'):
            ret += binascii.hexlify(self.hash).decode('ASCII')
        if hasattr(self, 'path'):
            ret += ', ' + self.path
        return ret

    def __repr__(self):
        return self.__str__()



from enum import IntEnum


class ETextureSetType(IntEnum):
    Invalid = 0x0,
    Color = 0x1,
    Alpha = 0x2,
    Normal = 0x3,
    Reflectance = 0x4,
    AO = 0x5,
    Roughness = 0x6,
    Height = 0x7,
    Mask = 0x8,
    Mask_Alpha = 0x9,
    Incandescence = 0xA,
    Translucency_Diffusion = 0xB,
    Translucency_Amount = 0xC,
    Misc_01 = 0xD,
    Count = 0xE
